/*  libatheos.so - the highlevel API library for AtheOS
 *  Copyright (C) 1999 - 2001  Kurt Skauen
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of version 2 of the GNU Library
 *  General Public License as published by the Free Software
 *  Foundation.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 *  MA 02111-1307, USA
 */

#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <dirent.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <sys/stat.h>

#include <atheos/time.h>

#include <gui/window.h>

#include <gui/stringview.h>

#include <gui/tableview.h>
#include <gui/requesters.h>
#include <gui/icon.h>
#include <gui/font.h>
#include <util/looper.h>
#include <util/message.h>
#include <storage/directory.h>
#include <storage/nodemonitor.h>


#include <list>
#include <set>
#include <cassert>

#include "directoryview.h"

using namespace os;

namespace os_priv
{
    enum Error_e { E_OK, E_SKIP, E_CANCEL, E_ALL };
    struct CopyFileParams_s
    {
		CopyFileParams_s( const std::vector<std::string> cDstPaths,
						const std::vector<std::string> cSrcPaths, const Messenger& cViewTarget ) :
				m_cDstPaths(cDstPaths), m_cSrcPaths(cSrcPaths), m_cViewTarget(cViewTarget) {}
  
		std::vector<std::string> m_cDstPaths;
		std::vector<std::string> m_cSrcPaths;
		Messenger				 m_cViewTarget;
    };

    struct DeleteFileParams_s
    {
		DeleteFileParams_s( const std::vector<std::string> cPaths, const Messenger& cViewTarget ) :
				m_cPaths(cPaths), m_cViewTarget(cViewTarget) {}
		std::vector<std::string> m_cPaths;
		Messenger				 m_cViewTarget;
    };

    class DirKeeper : public Looper
    {
    public:
		enum { M_CHANGE_DIR = 1, M_COPY_FILES, M_MOVE_FILES, M_DELETE_FILES, M_RENAME_FILES, M_ENTRIES_ADDED, M_ENTRIES_REMOVED, M_ENTRIES_UPDATED, M_LAYOUT_UPDATED };
		DirKeeper( const Messenger& cTarget, const std::string& cPath );
		virtual void HandleMessage( Message* pcMessage );
		virtual bool Idle();
		void Stop();
		status_t GetNode( os::String zPath, os::FSNode* pcNode, bool *pbBrokenLink );
		Directory* GetCurrentDir() { return( m_pcCurrentDir ); }
		bool IsReading() { return( m_eState == S_READDIR ); }
    private:
		void SendAddMsg( const String& cName, dev_t nDevice, ino_t nInode );
		void SendRemoveMsg( const String& cName, dev_t nDevice, ino_t nInode );
		void SendUpdateMsg( const String& cName, dev_t nDevice, ino_t nInode, bool bReloadIcon );
		
		enum state_t
			{ S_IDLE, S_READDIR };
		Messenger    m_cTarget;
		os::String	 m_cPath;
		Directory*   m_pcCurrentDir;
		NodeMonitor  m_cMonitor;
		state_t	     m_eState;
		int			 m_nPendingReplies;
		bool	     m_bWaitForAddReply;
		bool	     m_bWaitForRemoveReply;
		bool	     m_bWaitForUpdateReply;
		bool		 m_bLayoutNecessary;
		bool		 m_bWaitForLayoutReply;
		bigtime_t	 m_nLastLayout;
	
		struct MonitoringFileNode;
		struct FileNode
		{
	   		FileNode( const std::string& cName, const std::string& cPath, dev_t nDevice, ino_t nInode ) : m_cName( cName ), m_cPath( cPath ), m_nDevice( nDevice ), m_nInode( nInode )
			{
			}

	   		bool operator<(const FileNode& cNode ) const
			{
				return( m_nInode == cNode.m_nInode ? (m_nDevice < cNode.m_nDevice) : (m_nInode < cNode.m_nInode) );
	   		}
			bool operator<(const MonitoringFileNode& cNode ) const;
	    
	    	mutable std::string m_cName, m_cPath;
	    	dev_t	m_nDevice;
	    	ino_t	m_nInode;
		};

		struct MonitoringFileNode : public FileNode
		{
			MonitoringFileNode( const std::string& cName, dev_t nDevice, ino_t nInode ) : FileNode( cName, cName, nDevice, nInode )
			{
			}
	    	mutable NodeMonitor m_cMonitor;
		};

		std::set<MonitoringFileNode> m_cFileMap;
		std::set<FileNode> m_cAddMap;
		std::set<FileNode> m_cUpdateMap;
		std::set<FileNode> m_cRemoveMap;
    	};

    	bool DirKeeper::FileNode::operator<(const DirKeeper::MonitoringFileNode& cNode ) const
		{
			return( m_nInode == cNode.m_nInode ? (m_nDevice < cNode.m_nDevice) : (m_nInode < cNode.m_nInode) );
    	}
}

using namespace os_priv;

bool load_icon( const char* pzPath, void* pBuffer, bool bLarge )
{
	IconDir sDir;
	IconHeader sHeader;

	FILE* hFile = fopen( pzPath, "r" );
	if ( hFile == NULL )
	{
		printf( "Failed to open file %s\n", pzPath );
		return( false );
	}

	if ( fread( &sDir, sizeof( sDir ), 1, hFile ) != 1 )
		printf( "Failed to read icon dir\n" );

	if ( sDir.nIconMagic != ICON_MAGIC )
	{
		printf( "Files %s is not an icon\n", pzPath );
		return( false );
	}

	for ( int i = 0 ; i < sDir.nNumImages ; ++i )
	{
		if ( fread( &sHeader, sizeof( sHeader ), 1, hFile ) != 1 )
			printf( "Failed to read icon header\n" );

		if ( sHeader.nWidth == 32 )
		{
			if ( bLarge )
			{
				fread( pBuffer, 32*32*4, 1, hFile );
			}
			else
			{
				fseek( hFile, 32*32*4, SEEK_CUR );
			}
		}
		else if ( sHeader.nWidth == 16 )
		{
			if ( bLarge == false )
			{
				fread( pBuffer, 16*16*4, 1, hFile );
			}
			else
			{
				fseek( hFile, 16*16*4, SEEK_CUR );
			}
		}
	}

	fclose( hFile );
	return( true );
}

DirKeeper::DirKeeper( const Messenger& cTarget, const std::string& cPath ) : Looper( "dir_worker" ),  m_cTarget( cTarget )
{
	m_pcCurrentDir = NULL;
	try
	{
		m_cPath = cPath;
		m_pcCurrentDir = new Directory( cPath );
		m_pcCurrentDir->Rewind();
		m_cMonitor.SetTo( m_pcCurrentDir, NWATCH_DIR | NWATCH_NAME, this, this );
	}
	catch( std::exception& )
	{
	}
	m_eState = S_IDLE;
	m_nPendingReplies = 0;
	m_bLayoutNecessary = false;
	m_bWaitForLayoutReply = false;
	m_nLastLayout = get_system_time() - LAYOUT_TIME;
}

void DirKeeper::HandleMessage( Message* pcMessage )
{
	switch( pcMessage->GetCode() )
	{
		case M_CHANGE_DIR:
		{
			String cNewPath;
			pcMessage->FindString( "path", &cNewPath );
			try
			{
				Stop();
				Directory* pcNewDir = new Directory( cNewPath );
				delete m_pcCurrentDir;
				m_pcCurrentDir = pcNewDir;
				m_eState = S_READDIR;
				m_cTarget.SendMessage( DirectoryView::M_CLEAR );

				m_cFileMap.clear();
				m_cAddMap.clear();
				m_cRemoveMap.clear();
				m_cUpdateMap.clear();
				m_bWaitForAddReply    = false;
				m_bWaitForRemoveReply = false;
				m_bWaitForUpdateReply = false;
			}
			catch ( std::exception& )
			{
			}
			break;
		}
	case M_ENTRIES_ADDED:
		{
			m_bWaitForAddReply = false;
			if( m_cAddMap.empty() == false )
			{
				Message cMsg( DirectoryView::M_ADD_ENTRY );
				int nCount = 0;

				while( m_cAddMap.empty() == false )
				{
					std::set<FileNode>::iterator i = m_cAddMap.begin();

					try
					{
						FSNode cNode( *m_pcCurrentDir, (*i).m_cPath, O_RDONLY | O_NONBLOCK );
						struct stat sStat;

						cNode.GetStat( &sStat, false );









						cMsg.AddString( "name", (*i).m_cName );
						cMsg.AddString( "path", (*i).m_cPath );
						cMsg.AddData( "stat", T_ANY_TYPE, &sStat, sizeof( sStat ) );
					}
					catch( std::exception & )
					{
					}
					m_cAddMap.erase( i );
					if( ++nCount > 5 )
						break;
				}

				m_bWaitForAddReply = true;
				m_cTarget.SendMessage( &cMsg );
			}
			break;
		}
		case M_ENTRIES_UPDATED:
		{
			m_bWaitForUpdateReply = false;
			if ( m_cUpdateMap.empty() == false )
			{
				Message cMsg( DirectoryView::M_UPDATE_ENTRY );
				int nCount = 0;

				while ( m_cUpdateMap.empty() == false )
				{
					std::set<FileNode>::iterator i = m_cUpdateMap.begin();

					try
					{
						FSNode cNode( *m_pcCurrentDir, (*i).m_cName, O_RDONLY | O_NONBLOCK );
						struct stat sStat;
						cNode.GetStat( &sStat, false );
						uint32 anIcon[16*16];

						if ( S_ISDIR( sStat.st_mode ) == false )
						{
			    			load_icon( "/system/icons/file.icon", anIcon, false );
						}
						else
						{
			    			load_icon( "/system/icons/folder.icon", anIcon, false );
						}











						cMsg.AddString( "name", (*i).m_cName.c_str() );
						cMsg.AddData( "stat", T_ANY_TYPE, &sStat, sizeof( sStat ) );
						cMsg.AddData( "small_icon", T_ANY_TYPE, anIcon, 16*16*4 );

					} catch( std::exception& )
					{
					}
					m_cUpdateMap.erase( i );
					if ( ++nCount > 5 )
						break;
				}
				m_bWaitForUpdateReply = true;
				m_cTarget.SendMessage( &cMsg );
			}
			break;
		}
		case M_ENTRIES_REMOVED:
		{
			m_bWaitForRemoveReply = false;

			if ( m_cRemoveMap.empty() == false )
			{
				Message cMsg( DirectoryView::M_REMOVE_ENTRY );
				int nCount = 0;

				while ( m_cRemoveMap.empty() == false )
				{
					std::set<FileNode>::iterator i = m_cRemoveMap.begin();
					cMsg.AddInt32( "device", (*i).m_nDevice );
					cMsg.AddInt64( "node", (*i).m_nInode );
					m_cRemoveMap.erase( i );

					if ( ++nCount > 10 )
						break;
				}
				m_cTarget.SendMessage( &cMsg );
				m_bWaitForRemoveReply = true;
			}
			break;
		}
	case M_LAYOUT_UPDATED:
		{
			m_bWaitForLayoutReply = false;
			m_nLastLayout = get_system_time();
			break;
		}
	
	case M_NODE_MONITOR:
		{
			int32 nEvent = -1;

			pcMessage->FindInt32( "event", &nEvent );
			switch ( nEvent )
			{
				case NWEVENT_CREATED:
				{
					String cName;
					dev_t nDevice;
					int64 nInode;

					pcMessage->FindInt( "device", &nDevice );
					pcMessage->FindInt64( "node", &nInode );
					pcMessage->FindString( "name", &cName );
					SendAddMsg( cName, nDevice, nInode );
					break;
				}
				case NWEVENT_DELETED:
				{
					dev_t nDevice;
					int64 nInode;
					String cName;

					pcMessage->FindInt( "device", &nDevice );
					pcMessage->FindInt64( "node", &nInode );
					pcMessage->FindString( "name", &cName );
					SendRemoveMsg( cName, nDevice, nInode );
					break;
				}
				case NWEVENT_STAT_CHANGED:
				{
					dev_t nDevice;
					int64 nInode;
					String cName;

					pcMessage->FindInt( "device", &nDevice );
					pcMessage->FindInt64( "node", &nInode );
					pcMessage->FindString( "name", &cName );
					SendUpdateMsg( cName, nDevice, nInode, false );
					break;
				}
				case NWEVENT_ATTR_WRITTEN:
				case NWEVENT_ATTR_DELETED:
				{
				/* Not implemented yet in the kernel/glibc */
				break;
				}
				case NWEVENT_MOVED:
				{
					dev_t nDevice;
					int64 nOldDir;
					int64 nNewDir;
					int64 nInode;
					String cName;

					pcMessage->FindInt( "device", &nDevice );
					pcMessage->FindInt64( "old_dir", &nOldDir );
					pcMessage->FindInt64( "new_dir", &nNewDir );
					pcMessage->FindInt64( "node", &nInode );
					pcMessage->FindString( "name", &cName );


					if( nInode == int64 ( m_pcCurrentDir->GetInode() ) )
					{
						printf( "%Ld moved from %Ld to %Ld\n", nInode, nOldDir, nNewDir );
					}
					else
					{
						if( nOldDir == nNewDir )
						{
							SendUpdateMsg( cName, nDevice, nInode, true );
						}
						else
						{
							if( nOldDir == int64 ( m_pcCurrentDir->GetInode() ) )
							{
								SendRemoveMsg( cName, nDevice, nInode );
							}
							else
							{
								SendAddMsg( cName, nDevice, nInode );
							}
						}
					}
					break;
				}
			}
		}
	default:
		Looper::HandleMessage( pcMessage );
		break;
	}
}

status_t DirKeeper::GetNode( os::String zPath, os::FSNode* pcNode, bool* pbBrokenLink )
{	
	if( pcNode->SetTo( *m_pcCurrentDir, zPath, O_RDONLY | O_NONBLOCK ) >= 0 )
	{
		*pbBrokenLink = false;
			return( 0 );
	}
	
	if( pcNode->SetTo( *m_pcCurrentDir, zPath, O_RDONLY | O_NONBLOCK | O_NOTRAVERSE ) >= 0 )
	{
		*pbBrokenLink = true;		
			return( 0 );
	}
		
	return( -EIO );
}

void DirKeeper::SendUpdateMsg( const String& cName, dev_t nDevice, ino_t nInode, bool bReloadIcon )
{





	try
	{



		bool bBrokenLink;
		struct stat sStat;
		
		if( cName == "" )
		{
			bBrokenLink = false;
			sStat.st_dev = nDevice;
			sStat.st_ino = nInode;
		}
		else
		{
			FSNode cNode;
			if( GetNode( cName, &cNode, &bBrokenLink ) < 0 )
				return;
			cNode.GetStat( &sStat, false );
		}
			






		//FileNode cFileNode( cName, cName, sStat, bBrokenLink, bReloadIcon );
		
		Message cMsg( DirectoryView::M_UPDATE_ENTRY );
		cMsg.AddString( "name", cName );
		cMsg.AddString( "path", cName );
		cMsg.AddData( "stat", T_ANY_TYPE, &sStat, sizeof( sStat ) );
		cMsg.AddBool( "broken_link", bBrokenLink );
		cMsg.AddBool( "reload_icon", bReloadIcon );
			
		m_nPendingReplies++;
		m_cTarget.SendMessage( &cMsg );
	}
	catch( std::exception & )
	{
		m_cTarget.SendMessage( DirectoryView::M_REREAD );
	}
	m_bLayoutNecessary = true;
}

/**
 * Code Update: CVS SendAddMsg function from icondirectoryview
 *****************************************************************************/
void DirKeeper::SendAddMsg( const String& cName, dev_t nDevice, ino_t nInode )
{
	try
	{
		FSNode cNode;
		bool bBrokenLink = true;
		os::NodeMonitor* pcMonitor = NULL;
		struct stat sStat;
		memset( &sStat, 0, sizeof( sStat ) );
		
		if( GetNode( cName, &cNode, &bBrokenLink ) >= 0 )
		{
			cNode.GetStat( &sStat, false );
		}
		else
		{
			if( nDevice == -1 && nInode == -1 )
				return;

			sStat.st_dev = nDevice;
			sStat.st_ino = nInode;
		}
		if( !bBrokenLink )
		{
			pcMonitor = new os::NodeMonitor( &cNode, NWATCH_STAT | NWATCH_ATTR, this, this );
		}
		Message cMsg( DirectoryView::M_ADD_ENTRY );
		cMsg.AddString( "name", cName );
		cMsg.AddString( "path", cName );
		cMsg.AddData( "stat", T_ANY_TYPE, &sStat, sizeof( sStat ) );
		cMsg.AddBool( "broken_link", bBrokenLink );
		cMsg.AddPointer( "node_monitor", pcMonitor );
		
		m_nPendingReplies++;
		m_cTarget.SendMessage( &cMsg );
	}
	catch( ... /*std::exception& */  )
	{

	}
	m_bLayoutNecessary = true;
}

void DirKeeper::SendRemoveMsg( const String& cName, dev_t nDevice, ino_t nInode )
{
	struct stat sStat;
	sStat.st_dev = nDevice;
	sStat.st_ino = nInode;
	
	

















	Message cMsg( DirectoryView::M_REMOVE_ENTRY );
	cMsg.AddInt32( "device", nDevice );
	cMsg.AddInt64( "node", nInode );
	
	m_nPendingReplies++;
	m_cTarget.SendMessage( &cMsg );
	m_bLayoutNecessary = true;
}






void DirKeeper::Stop()
{
	Lock();
	m_eState = S_IDLE;
	m_bLayoutNecessary = false;
	m_bWaitForLayoutReply = false;
	m_cMonitor.Unset();
	
	m_nPendingReplies = 0;
	
	/* Make sure we delete all pending messages */
	SpoolMessages();	// Copy all messages from the message port to the message queue.
	MessageQueue *pcQueue = GetMessageQueue();
	MessageQueue cTmp;
	Message* pcMsg;
	pcQueue->Lock();
	while( ( pcMsg = pcQueue->NextMessage() ) != NULL )
	{
		switch( pcMsg->GetCode() )
		{
			case M_ENTRIES_ADDED:
			case M_ENTRIES_UPDATED:
			case M_ENTRIES_REMOVED:
			case M_NODE_MONITOR:
				break;
			default:
				cTmp.AddMessage( pcMsg );
		}
	}
	
	while( ( pcMsg = cTmp.NextMessage() ) != NULL )
	{
		pcQueue->AddMessage( pcMsg );
	}
	
	pcQueue->Unlock();
	Unlock();
}

bool DirKeeper::Idle()
{
	switch( m_eState )
	{
		case S_READDIR:
		{
			if ( m_pcCurrentDir == NULL )
			{
				m_eState = S_IDLE;
				return( false );
			}

			/* Show files */
			String cName;
			if( m_bLayoutNecessary )
			{
				if( m_nPendingReplies > 0 )
				{
					if( get_system_time() - m_nLastLayout > LAYOUT_TIME && !m_bWaitForLayoutReply )
					{
						/* Update regulary */
						Message cMsg( DirectoryView::M_LAYOUT );
						m_cTarget.SendMessage( &cMsg );
						m_bWaitForLayoutReply = true;
						m_bLayoutNecessary = false;
						m_nLastLayout = get_system_time();
					}
				}
			}
			
			if( m_nPendingReplies > 5 )
				return( false );
				
			if( m_pcCurrentDir->GetNextEntry( &cName ) == 1 )
			{
				if( !( cName == "." ) && !( cName == ".." ) )
				{
					SendAddMsg( cName, -1, -1 );
				}
				return ( true );
			}
			else
			{
				m_eState = S_IDLE;
				return( false );
			}
			break;
		}
		case S_IDLE:
		if( m_bLayoutNecessary )
		{
			if( m_nPendingReplies > 0 )
				return( false );

			/* Final layout */			
			Message cMsg( DirectoryView::M_LAYOUT );
			m_cTarget.SendMessage( &cMsg );
			m_nLastLayout = get_system_time();
			m_bWaitForLayoutReply = true;
			m_bLayoutNecessary = false;
		}
		return ( false );
	}

	return( false );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

int ReadFileDlg( int nFile, void* pBuffer, int nSize, const char* pzPath, Error_e* peError )
{
retry:
	int nLen = read( nFile, pBuffer, nSize );

	if ( nLen >= 0 )
	{
		*peError = E_OK;
		return( nLen );
	}
	else
	{
		if ( errno == EINTR )
			goto retry;
	}
	std::string cMsg("Failed to read from source file: ");
	cMsg += pzPath;
	cMsg += "\nError: ";
	cMsg += strerror( errno );
	Alert* pcAlert = new Alert( "Error:",
								cMsg.c_str(), 0,
								"Skip", "Retry", "Cancel" );
	switch( pcAlert->Go() )
	{
		case 0:
			*peError = E_SKIP;
			break;
		case 1:
			goto retry;
		case 2:
			*peError = E_CANCEL;
			break;
		default:
			*peError = E_SKIP;
			break;
	}

	return( nLen );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

int WriteFileDlg( int nFile, const void* pBuffer, int nSize, const char* pzPath, Error_e* peError )
{
retry:
	int nLen = write( nFile, pBuffer, nSize );
	if ( nLen >= 0 )
	{
		*peError = E_OK;
		return( nLen );
	}
	else
	{
		if ( errno == EINTR )
			goto retry;
	}

	std::string cMsg("Failed to write to destination file: ");
	cMsg += pzPath;
	cMsg += "\nError: ";
	cMsg += strerror( errno );
	Alert* pcAlert = new Alert( "Error:",
								cMsg.c_str(), 0,
								"Skip", "Retry", "Cancel" );
	switch( pcAlert->Go() )
	{
		case 0:
			*peError = E_SKIP;
			break;
		case 1:
			goto retry;
		case 2:
			*peError = E_CANCEL;
			break;
		default:
			*peError = E_SKIP;
			break;
	}

	return( nLen );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

bool CopyFile( const char* pzDst, const char* pzSrc,
               bool* pbReplaceFiles, bool* pbDontOverwrite, ProgressRequester* pcProgress )
{
	struct stat sSrcStat;
	struct stat sDstStat;
	bool        bDirExists = false;

	if ( pcProgress->DoCancel() )
		return( false );

	while ( lstat( pzSrc, &sSrcStat ) < 0 )
	{
		std::string cMsg("Failed to open source: ");
		cMsg += pzSrc;
		cMsg += "\nError: ";
		cMsg += strerror( errno );
		Alert* pcAlert = new Alert( "Error:",
									cMsg.c_str(), 0,
									"Skip", "Retry", "Cancel" );
		switch( pcAlert->Go() )
		{
			case 0:
		        return( true );
			case 1:
		        break;
			case 2:
		        return( false );
			default:
		       return( true );
		}
	}

	bool bLoop = true;
	while( bLoop && lstat( pzDst, &sDstStat ) >= 0 )
	{
		if ( sSrcStat.st_dev == sDstStat.st_dev && sSrcStat.st_ino == sDstStat.st_ino )
		{
			Alert* pcAlert = new Alert( "Error", "Source and destination are the same, can't copy.", 0, "OK");
			pcAlert->Go( NULL );
			return( false );
		}

		if ( S_ISDIR( sDstStat.st_mode ) )
		{
			if ( S_ISDIR( sSrcStat.st_mode ) == false )
			{
				std::string cMsg("Can't replace directory  ");
				cMsg += pzDst;
				cMsg += " with a file\n";
				Alert* pcAlert = new Alert( "Error",
											  cMsg.c_str(), 0,
											  "Skip", "Retry", "Cancel" );
				switch( pcAlert->Go() )
				{
					case 0:
						return( true );
					case 1:
						break;
					case 2:
						return( false );
					default:
						return( true );
				}
			}
			else
			{
				bDirExists = true;
				break;
			}
		}
		else
		{
			if ( *pbDontOverwrite )
				return( true );
			
			if ( *pbReplaceFiles )
			{
				unlink( pzDst );
				break;
			}

			std::string cMsg("The destination file: ");
			cMsg += pzDst;
			cMsg += "already exists\nWould you like to replace it?";
			Alert* pcAlert = new Alert( "Error:",
										cMsg.c_str(), 0,
										"Skip", "Replace", "Cancel" );
			switch( pcAlert->Go() )
			{
				case 0:        // Skip
					return( true );
					/*
				case 1: // Skip all
					*pbDontOverwrite = true;
					return( true );
					*/
				case 1:        // Replace
					bLoop = false;
					unlink( pzDst );
					break;
					/*
				case 3:        // Replace all
					*pbReplaceFiles = true;
					bLoop = false;
					unlink( pzDst );
					break;
					*/
				case 2:        // Cancel
					return( false );
				default: // Bad
					return( true );
			}
		}
	}

	if ( S_ISDIR( sSrcStat.st_mode ) )
	{
		pcProgress->Lock();
		pcProgress->SetPathName( pzDst );
		pcProgress->Unlock();

		if ( bDirExists == false )
		{
			while ( mkdir( pzDst, sSrcStat.st_mode ) < 0 )
			{
				std::string cMsg("Failed to create directory: ");
				cMsg += pzDst;
				cMsg += "\nError: ";
				cMsg += strerror( errno );
				Alert* pcAlert = new Alert( "Error",
											  cMsg.c_str(), 0,
											  "Skip", "Retry", "Cancel" );
				switch( pcAlert->Go() )
				{
					case 0:
						return( true );
					case 1:
						break;
					case 2:
						return( false );
					default:
						return( true );
				}
			}
		}
		DIR* pDir;

		for (;;)
		{
			pDir = opendir( pzSrc );
			if ( pDir != NULL )
				break;

			std::string cMsg("Failed to open directory: ");
			cMsg += pzSrc;
			cMsg += "\nError: ";
			cMsg += strerror( errno );
			Alert* pcAlert = new Alert( "Error",
										  cMsg.c_str(), 0,
										  "Skip", "Retry", "Cancel" );
			switch( pcAlert->Go() )
			{
				case 0:
					return( true );
				case 1:
					break;
				case 2:
					return( false );
				default:
					return( true );
			}
		}

		struct dirent* psEntry;

		while( (psEntry = readdir( pDir )) != NULL )
		{
			if ( strcmp( psEntry->d_name, "." ) == 0 || strcmp( psEntry->d_name, ".." ) == 0 )
				continue;

			Path cSrcPath( pzSrc );
			Path cDstPath( pzDst );
			cSrcPath.Append( psEntry->d_name );
			cDstPath.Append( psEntry->d_name );

			if ( CopyFile( cDstPath.GetPath().c_str(), cSrcPath.GetPath().c_str(),
						pbReplaceFiles, pbDontOverwrite, pcProgress ) == false )
			{
				closedir( pDir );
				return( false );
			}
		}
		closedir( pDir );
	}
	else if ( S_ISLNK( sSrcStat.st_mode ) )
	{
		pcProgress->Lock();
		pcProgress->SetFileName( Path( pzDst ).GetLeaf() );
		pcProgress->Unlock();
		printf( "Copy link %s to %s\n", pzSrc, pzDst );
	}
	else
	{
		pcProgress->Lock();
		pcProgress->SetFileName( Path( pzDst ).GetLeaf() );
		pcProgress->Unlock();

		int nSrcFile = -1;
		int nDstFile = -1;

		for (;;)
		{
			nSrcFile = open( pzSrc, O_RDONLY );
			if ( nSrcFile >= 0 )
				break;

			std::string cMsg("Failed to open source file: ");
			cMsg += pzSrc;
			cMsg += "\nError: ";
			cMsg += strerror( errno );
			Alert* pcAlert = new Alert( "Error:",
										  cMsg.c_str(), 0,
										  "Skip", "Retry", "Cancel" );
			switch( pcAlert->Go() )
			{
				case 0:
					return( true );
				case 1:
					break;
				case 2:
					return( false );
				default:
					return( true );
			}
		}

		for (;;)
		{
			nDstFile = open( pzDst, O_WRONLY | O_CREAT | O_TRUNC, sSrcStat.st_mode );
			if ( nDstFile >= 0 )
				break;

			std::string cMsg("Failed to create destination file: ");
			cMsg += pzDst;
			cMsg += "\nError: ";
			cMsg += strerror( errno );
			Alert* pcAlert = new Alert( "Error",
										cMsg.c_str(), 0,
										"Skip", "Retry", "Cancel" );
			switch( pcAlert->Go() )
			{
				case 0:
					close( nSrcFile );
					return( true );
				case 1:
					break;
				case 2:
					close( nSrcFile );
					return( false );
				default:
					close( nSrcFile );
					return( true );
			}
		}

		int nTotLen = 0;

		for (;;)
		{
			char anBuffer[4096];
			int  nLen;
			Error_e eError;

			if ( pcProgress->DoSkip() )
			{
				close( nSrcFile );
				close( nDstFile );
				unlink( pzDst );
				return( true );
			}
			
			for (;;)
			{
				nLen = ReadFileDlg( nSrcFile, anBuffer, 4096, pzSrc, &eError );
				if ( nLen >= 0 )
					break;

				close( nSrcFile );
				close( nDstFile );
				unlink( pzDst );

				if ( eError == E_SKIP )
				{
					return( true );
				}
				else if ( eError == E_CANCEL )
				{
					return( false );
				}
				else
				{
					assert( !"Invalid return code from ReadFileDlg()" );
					return( false );
				}
			}
			
			for (;;)
			{
				int nError = WriteFileDlg( nDstFile, anBuffer, nLen, pzDst, &eError );
				if ( nError >= 0 )
					break;

				close( nSrcFile );
				close( nDstFile );
				unlink( pzDst );

				if ( eError == E_SKIP )
				{
					return( true );
				}
				else if ( eError == E_CANCEL )
				{
					return( false );
				}
				else
				{
					assert( !"Invalid return code from WriteFileDlg()" );
					return( false );
				}
			}
			nTotLen += nLen;
			if ( nLen < 4096 )
				break;
		}
		close( nSrcFile );
		close( nDstFile );

		if ( nTotLen != sSrcStat.st_size )
		{
			std::string cMsg("Mismatch between number of bytes read (");
			cMsg += nTotLen;
			cMsg += ")\n";
			cMsg += "and file size reported by file system (";
			cMsg += sSrcStat.st_size;
			cMsg += ")!\n";
			cMsg += "It seems like we might end up corrupting the destination file\n";
			Alert* pcAlert = new Alert( "Warning", cMsg.c_str(), 0, "Sorry" );
			pcAlert->Go();
		}
	}
	return( true );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

bool DeleteFile( const char* pzPath, ProgressRequester* pcProgress )
{
	struct stat sStat;

	if ( pcProgress->DoCancel() )
		return( false );

	while ( lstat( pzPath, &sStat ) < 0 )
	{
		std::string cMsg("Failed to stat file: ");
		cMsg += pzPath;
		cMsg += "\nError: ";
		cMsg += strerror( errno );
		Alert* pcAlert = new Alert( "Error:",
									cMsg.c_str(), 0,
									"Skip", "Retry", "Cancel" );
		switch( pcAlert->Go() )
		{
			case 0:
				return( true );
			case 1:
				break;
			case 2:
				return( false );
			default:
				return( true );
		}
	}

	if ( S_ISDIR( sStat.st_mode ) )
	{
		pcProgress->Lock();
		pcProgress->SetPathName( pzPath );
		pcProgress->Unlock();

		DIR* pDir;
		for (;;)
		{
			pDir = opendir( pzPath );
			if ( pDir != NULL )
				break;

			std::string cMsg("Failed to open directory: ");
			cMsg += pzPath;
			cMsg += "\nError: ";
			cMsg += strerror( errno );
			Alert* pcAlert = new Alert( "Error:",
										cMsg.c_str(), 0,
										"Skip", "Retry", "Cancel" );
			switch( pcAlert->Go() )
			{
				case 0:
					return( true );
				case 1:
					break;
				case 2:
					return( false );
				default:
					return( true );
			}
		}
		struct dirent* psEntry;

		std::list<std::string> cFileList;

		while( (psEntry = readdir( pDir )) != NULL )
		{
			if ( strcmp( psEntry->d_name, "." ) == 0 || strcmp( psEntry->d_name, ".." ) == 0 )
				continue;

			Path cPath( pzPath );
			cPath.Append( psEntry->d_name );
			cFileList.push_back( cPath.GetPath() );
		}
		closedir( pDir );

		while( cFileList.empty() == false )
		{
			std::list<std::string>::iterator i = cFileList.begin();
			if ( DeleteFile( (*i).c_str(), pcProgress ) == false )
				return( false );

			cFileList.erase( i );
		}

		while( rmdir( pzPath ) < 0 )
		{
			std::string cMsg("Failed to delete directory: ");
			cMsg += pzPath;
			cMsg += "\nError: ";
			cMsg += strerror( errno );
			Alert* pcAlert = new Alert( "Error:",
										cMsg.c_str(), 0,
										"Skip", "Retry", "Cancel" );
			switch( pcAlert->Go() )
			{
				case 0:
					return( true );
				case 1:
					break;
				case 2:
					return( false );
				default:
					return( true );
			}
		}
	}
	else
	{
		pcProgress->Lock();
		pcProgress->SetFileName( Path( pzPath ).GetLeaf() );
		pcProgress->Unlock();

		while( unlink( pzPath ) < 0 )
		{
			std::string cMsg("Failed to delete file: ");
			cMsg += pzPath;
			cMsg += "\nError: ";
			cMsg += strerror( errno );
			Alert* pcAlert = new Alert( "Error",
										  cMsg.c_str(), 0,
										  "Skip", "Retry", "Cancel" );
			switch( pcAlert->Go() )
			{
				case 0:
					return( true );
				case 1:
					break;
				case 2:
					return( false );
				default:
					return( true );
			}
		}
	}
	return( true );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

int32 DelFileThread( void* pData )
{
	DeleteFileParams_s* psArgs = (DeleteFileParams_s*) pData;

	std::string cMsg;
	if ( psArgs->m_cPaths.size() == 1 )
	{
		struct stat sStat;
		if ( lstat( psArgs->m_cPaths[0].c_str(), &sStat ) < 0 )
		{
			cMsg += "Failed to open file ";
			cMsg += psArgs->m_cPaths[0].c_str();
			cMsg += "\nError: ";
			cMsg += strerror( errno );
			Alert* pcAlert = new Alert( "Error", cMsg.c_str(), 0, "Sorry" );
			pcAlert->Go();
			delete psArgs;
			return( 1 );
		}
		if ( S_ISDIR( sStat.st_mode ) )
		{
			cMsg += "Are you sure you want to delete\n'the directory \'";
			cMsg += Path( psArgs->m_cPaths[0].c_str() ).GetLeaf().c_str();
			cMsg += "'\'\nand all its sub-directories?";
		}
		else
		{
			cMsg += ("Are you sure you want to delete the file\n");
			cMsg += "'\'";
			cMsg += Path( psArgs->m_cPaths[0].c_str() ).GetLeaf().c_str();
			cMsg += "'\'\n";
		}
	}
	else
	{
		cMsg += ("Are you sure you want to delete those ");
		cMsg += psArgs->m_cPaths.size();
		cMsg += " entries?";
	}

	Alert* pcAlert = new Alert( "Verify file deletion", cMsg.c_str(), 0, "Cancel", "Delete" );

	int nButton = pcAlert->Go();

	if ( nButton != 1 )
	{
		delete psArgs;
		return( 1 );
	}

	ProgressRequester* pcProgress = new ProgressRequester( Rect( 50, 20, 350, 150 ),
									"progress_window", "Delete file:", false );

	for ( uint i = 0 ; i < psArgs->m_cPaths.size() ; ++i )
	{
		if ( DeleteFile( psArgs->m_cPaths[i].c_str(), pcProgress ) == false )
			break;

		Message cMsg( 125 );
		psArgs->m_cViewTarget.SendMessage( &cMsg );
	}

	pcProgress->Quit();

	delete psArgs;
	return( 0 );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void StartFileDelete( const std::vector<std::string>& cPaths, const Messenger& cViewTarget )
{
	DeleteFileParams_s* psParams = new DeleteFileParams_s( cPaths, cViewTarget );
	thread_id hTread = spawn_thread( "delete_file_thread", (void*)DelFileThread, 0, 0, psParams );

	if ( hTread >= 0 )
		resume_thread( hTread );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

int32 CopyFileThread( void* pData )
{
	CopyFileParams_s* psParams = (CopyFileParams_s*) pData;

	ProgressRequester* pcProgress = new ProgressRequester( Rect( 50, 20, 350, 150 ),
									"progress_window", "Copy files:", true );

	for ( uint i = 0 ; i < psParams->m_cSrcPaths.size() ; ++i )
	{
		bool bReplaceFiles  = false;
		bool bDontOverwrite = false;

		if ( CopyFile( psParams->m_cDstPaths[i].c_str(), psParams->m_cSrcPaths[i].c_str(),
					&bReplaceFiles, &bDontOverwrite, pcProgress ) == false )
			break;

		Message cMsg( 125 );
		psParams->m_cViewTarget.SendMessage( &cMsg );
	}
	pcProgress->Quit();

	delete psParams;
	return( 0 );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void StartFileCopy( const std::vector<std::string>& cDstPaths,
                    const std::vector<std::string>& cSrcPaths, const Messenger& cViewTarget )
{
	CopyFileParams_s* psParams = new CopyFileParams_s( cDstPaths, cSrcPaths, cViewTarget );
	thread_id hTread = spawn_thread( "copy_file_thread", (void*)CopyFileThread, 0, 0, psParams );

	if ( hTread >= 0 )
		resume_thread( hTread );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

DirectoryView::State::State( ListView* pcView, const char* pzPath ) : m_cPath( pzPath )
{
	for ( uint i = 0 ; i < pcView->GetRowCount() ; ++i )
	{
		if ( pcView->IsSelected( i ) )
		{
			FileRow* pcRow = dynamic_cast<FileRow*>(pcView->GetRow(i));
			if ( pcRow != NULL )
				m_cSelection.push_back( pcRow->GetName() );
		}
	}
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

DirectoryView::DirectoryView( const Rect& cFrame, const std::string& cPath, uint32 nModeFlags, uint32 nResizeMask, uint32 nViewFlags ) :
    ListView( cFrame, "_list_view", nModeFlags, nResizeMask, nViewFlags ), m_cPath( cPath.c_str() )
{
	m_pcCurReadDirSession = NULL;
	m_bLocked = false;

	InsertColumn( "", 16, 0 );  // Icon
	InsertColumn( "Name", 200 );
	InsertColumn( "Size", 90 );
	InsertColumn( "Attr", 90 );
	InsertColumn( "Date", 100 );
	InsertColumn( "Time", 70 );

	m_pcDirChangeMsg = NULL;

	m_nLastKeyDownTime = 0;
	m_pcIconBitmap = new Bitmap( 16, 16, CS_RGB32 );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

DirectoryView::~DirectoryView()
{
	delete m_pcIconBitmap;
}

void DirectoryView::AttachedToWindow()
{


	ListView::AttachedToWindow();
	m_pcDirKeeper = new DirKeeper( Messenger( this ), m_cPath.GetPath() );
	m_pcDirKeeper->Run();
	ReRead();

}

void DirectoryView::DetachedFromWindow()
{
	if ( m_pcDirKeeper )
	{
		m_pcDirKeeper->Quit();
		m_pcDirKeeper = NULL;
	}
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void DirectoryView::KeyDown( const char* pzString, const char* pzRawString, uint32 nQualifiers )
{
	char nChar = pzString[0];

	if ( isprint( nChar ) )
	{
		bigtime_t nTime = get_system_time();

		if ( nTime < m_nLastKeyDownTime + 1000000 )
		{
			m_cSearchString.append( &nChar, 1 );
		}
		else
		{
			m_cSearchString = std::string( &nChar, 1 );
		}
		
		for ( uint i = 0 ; i < GetRowCount() ; ++i )
		{
			FileRow* pcRow = dynamic_cast<FileRow*>(GetRow(i));
			if ( pcRow != NULL )
			{
				// FIXME: gcc 3.0.4 doesn't like argument 2
				//			someone please tell me if my substitute function is equivalent
				
#if __GNUC__ < 3
				if ( m_cSearchString.compare( pcRow->m_cName, 0UL, m_cSearchString.size() ) == 0 )
#else
				if ( m_cSearchString.compare(0, m_cSearchString.size(), pcRow->m_cName, 0, m_cSearchString.size() ) == 0)
#endif
				{
					Select( i );
					MakeVisible( i, false );
					break;
				}
			}
		}
		m_nLastKeyDownTime = nTime;
	}
	else
	{
		switch( pzString[0] )
		{
			case VK_DELETE:
			{
				std::vector<std::string> cPaths;
				for ( int i = GetFirstSelected() ; i <= GetLastSelected() ; ++i )
				{
					if ( IsSelected( i ) )
					{
						FileRow* pcRow = dynamic_cast<FileRow*>(GetRow(i));
						if ( pcRow != NULL )
						{
							Path cPath = m_cPath;
							cPath.Append( pcRow->m_cName.c_str() );
							cPaths.push_back( cPath.GetPath() );
						}
					}
				}
				StartFileDelete( cPaths, Messenger( this ) );
				break;
			}

			case VK_BACKSPACE:

				if( m_bLocked)
					return( ListView::KeyDown( pzString, pzRawString, nQualifiers ) );

				m_cPath.Append( ".." );
				ReRead();
				PopState();
				DirChanged( m_cPath.GetPath() );
				break;
			
			case VK_FUNCTION_KEY:
			{
				Looper* pcLooper = GetLooper();
				assert( pcLooper != NULL );
				Message* pcMsg = pcLooper->GetCurrentMessage();
				assert( pcMsg != NULL );
				int32 nKeyCode;

				if ( pcMsg->FindInt32( "_raw_key", &nKeyCode ) != 0 )
					return;

				switch( nKeyCode )
				{
					case 6: // F5
						ReRead();
						break;
				}
				break;
			}
			default:
				ListView::KeyDown(pzString, pzRawString, nQualifiers);
				break;
		}
	}
}

void DirectoryView::PopupMenu( const Point &cPosition )
{
//	if( NULL == pcMenu )
//	{
		pcMenu = new Menu( Rect(), "popup_menu", ITEMS_IN_COLUMN );
		pcMenu->AddItem( "Cut", new Message( M_CUT ), "CTRL+X" );
		pcMenu->AddItem( new MenuSeparator );
		m_pcCopy = new MenuItem( "Copy", new Message( M_COPY ), "CTRL+C" );
		m_pcCopy->SetEnable( m_bSelect );
		pcMenu->AddItem( m_pcCopy );
		pcMenu->AddItem( "Paste", new Message( M_PASTE ), "CTRL+V" );
		pcMenu->SetTargetForItems( this );
//	}
		pcMenu->Open( ConvertToScreen( cPosition ) );
}

int32 DirectoryView::ReadDirectory( void* pData )
{
	ReadDirParam* pcParam = (ReadDirParam*) pData;
	DirectoryView* pcView = pcParam->m_pcView;

	Window* pcWnd = pcView->GetWindow();

	if ( pcWnd == NULL )
		return(0);

	pcWnd->Lock();
	pcView->Clear();
	pcWnd->Unlock();

	DIR* pDir = opendir( pcView->GetPath().c_str() );
	if ( pDir == NULL )
	{
		dbprintf( "Error: DirectoryView::ReadDirectory() Failed to open %s\n", pcView->GetPath().c_str() );
		goto error;
	}
	struct dirent* psEnt;

	while( pcParam->m_bCancel == false && (psEnt = readdir( pDir ) ) != NULL )
	{
		struct stat sStat;
		if ( strcmp( psEnt->d_name, "." ) == 0 /*|| strcmp( psEnt->d_name, ".." ) == 0*/ )
			continue;

		Path cFilePath( pcView->GetPath() );
		cFilePath.Append( psEnt->d_name );

		stat( cFilePath.GetPath().c_str(), &sStat );

		FileRow* pcRow = new FileRow( pcView->m_pcIconBitmap, psEnt->d_name, sStat );

		if ( S_ISDIR( pcRow->m_sStat.st_mode ) == false )
		{
		    load_icon( "/system/icons/file.icon", pcRow->m_anIcon, false );
		}
		else
		{
		    load_icon( "/system/icons/folder.icon", pcRow->m_anIcon, false );
		}

		pcWnd->Lock();
		pcView->InsertRow( pcRow );
		pcWnd->Unlock();
	}
	closedir( pDir );

error:
	pcWnd->Lock();
	pcView->Sort();

	if ( pcView->m_pcCurReadDirSession == pcParam )
		pcView->m_pcCurReadDirSession = NULL;

	pcWnd->Unlock();
	delete pcParam;

	return( 0 );
}

/** Rereads the directory content.
 * \par Description:
 * Rereads the directory content. Necessary after adding the view to a window
 * or after a SetPath() call.
 * \par Note:
 * Changes in the directory will automatically cause an update.
 * \author	Arno Klenke (arno_klenke@yahoo.de)
 *****************************************************************************/
void DirectoryView::ReRead()
{
	if( m_pcDirKeeper )
	{
//		m_pcDirKeeper->Lock(); // TODO: latches up the app
		m_pcDirKeeper->Stop();
	}

	Clear();

	/* Make sure we delete all pending messages from the dirkeeper thread */
	if( GetLooper() )
	{
		os::Looper* pcLooper = GetLooper();
		pcLooper->SpoolMessages();	// Copy all messages from the message port to the message queue.
		MessageQueue *pcQueue = pcLooper->GetMessageQueue();
		MessageQueue cTmp;
		Message *pcMsg;

		pcQueue->Lock();

		while( ( pcMsg = pcQueue->NextMessage() ) != NULL )
		{
			switch( pcMsg->GetCode() )
			{
				case M_ADD_ENTRY:
				case M_UPDATE_ENTRY:
				case M_REMOVE_ENTRY:
				case M_LAYOUT:
					break;
				default:
					cTmp.AddMessage( pcMsg );
			}
		}
		
		while( ( pcMsg = cTmp.NextMessage() ) != NULL )
			pcQueue->AddMessage( pcMsg );

		pcQueue->Unlock();
	}

	if ( m_pcCurReadDirSession != NULL )
        m_pcCurReadDirSession->m_bCancel = true;

    m_pcCurReadDirSession = new ReadDirParam( this );

	ReadDirectory(m_pcCurReadDirSession);
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

std::string DirectoryView::GetPath() const
{
	return( std::string( m_cPath.GetPath() ) );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void DirectoryView::SetPath( const std::string& cPath )
{
	while( m_cStack.size() > 0 )
	{
		m_cStack.pop();
	}
	m_cPath.SetTo( cPath.c_str() );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void DirectoryView::Invoked( int nFirstRow, int nLastRow )
{
	if ( nFirstRow != nLastRow )
	{
		ListView::Invoked( nFirstRow, nLastRow );
		return;
	}

	FileRow* pcRow = dynamic_cast<FileRow*>(GetRow( nFirstRow ));
	if ( pcRow == NULL )
	{
		ListView::Invoked( nFirstRow, nLastRow );
		return;
	}
	if ( S_ISDIR( pcRow->m_sStat.st_mode ) == false )
	{
		ListView::Invoked( nFirstRow, nLastRow );
		return;
	}

	if ( strcmp( pcRow->m_cName.c_str(), ".." ) == 0 )
	{
		m_cPath.Append( ".." );
		PopState();
	}
	else
	{
		m_cStack.push( State( this, m_cPath.GetPath().c_str() ) );
		m_cPath.Append( pcRow->m_cName.c_str() );
	}

	ReRead();
	DirChanged( m_cPath.GetPath() );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void DirectoryView::PopState()
{
	if ( m_cStack.empty() == false )
	{
		State& cState = m_cStack.top();

		for ( uint i = 0 ; i < cState.m_cSelection.size() ; ++i )
		{
			for ( uint j = 0 ; j < GetRowCount() ; ++j )
			{
				FileRow* pcRow = dynamic_cast<FileRow*>(GetRow(j));
				if ( pcRow != NULL )
				{
					if ( cState.m_cSelection[i] == pcRow->m_cName )
					{
						Select( j );
						MakeVisible( j );
						break;
					}
				}
			}
		}
		m_cStack.pop();
	}
}

/** Sets whether the current directory is locked.
 * \par Description:
 * Sets whether the current directory is locked. When the user tries to access
 * another directory, the message set by SetDirChangeMsg() is sent, but the 
 * directory is not changed.
 * \param bLocked - Enables/Disables this feature.
 * \author	Arno Klenke (arno_klenke@yahoo.de)
 *****************************************************************************/
void DirectoryView::SetDirectoryLocked( bool bLocked )
{
	m_bLocked = bLocked;
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void DirectoryView::SetDirChangeMsg( Message* pcMsg )
{
	delete m_pcDirChangeMsg;
	m_pcDirChangeMsg = pcMsg;
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void DirectoryView::DirChanged( const std::string& cNewPath )
{
	if( m_pcDirChangeMsg != NULL )
	{
		os::Message cMsg( *m_pcDirChangeMsg );
		cMsg.AddString( "file/path", cNewPath );
		Invoke( &cMsg );
	}
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void DirectoryView::DragSelection( Point cStartPoint )
{
	int nFirstSel = GetFirstSelected();
	int nLastSel  = GetLastSelected();
	int nNumRows  = nLastSel - nFirstSel + 1;
	Point cPos = cStartPoint;

	float vRowHeight = GetRow( 0 )->GetHeight( this ) + 3.0f;
	float vSelWidth = 0;

	Message cData(1234);
	for ( int i = nFirstSel ; i <= nLastSel ; ++i )
	{
		FileRow* pcRow = dynamic_cast<FileRow*>(GetRow(i));
		if ( pcRow != NULL && IsSelected( i ) )
		{
			float vLen = GetStringWidth( pcRow->m_cName.c_str() );
			if ( vLen > vSelWidth )
				vSelWidth = vLen;

			Path cPath = m_cPath;
			cPath.Append( pcRow->m_cName.c_str() );
			cData.AddString( "file/path", cPath.GetPath() );
		}
	}

	vSelWidth += 18.0f;
	Rect cSelRect( 0, 0, vSelWidth - 1.0f, vRowHeight * nNumRows - 1 );
	Point cHotSpot( cPos.x, cPos.y - GetRowPos( nFirstSel ) );

	if ( nLastSel - nFirstSel < 4 )
	{
		Bitmap cImage( cSelRect.Width()+1.0f, cSelRect.Height()+1.0f, CS_RGB32, Bitmap::ACCEPT_VIEWS );
		View* pcView = new View( cSelRect, "", CF_FOLLOW_LEFT | CF_FOLLOW_TOP, WID_WILL_DRAW );
		cImage.AddChild( pcView );
		pcView->SetFgColor( 255, 255, 255, 255 );
		pcView->FillRect( cSelRect );

		Rect cBmRect( 0, 0, 15, 15 );
		Rect cTextRect( 18, 0, vSelWidth - 1.0f, vRowHeight - 1.0f );

		pcView->SetBgColor( 100, 100, 100 );
		for ( int i = nFirstSel ; i <= nLastSel ; ++i )
		{
			FileRow* pcRow = dynamic_cast<FileRow*>(GetRow(i));
			if ( pcRow != NULL && IsSelected( i ) )
			{
				pcView->SetDrawingMode( DM_OVER );
				pcRow->Paint( cBmRect, pcView, 0, false, false, false );
				pcView->SetDrawingMode( DM_COPY );
				pcRow->Paint( cTextRect, pcView, 6, false, false, false );
			}
			cBmRect += Point( 0.0f, vRowHeight );
			cTextRect += Point( 0.0f, vRowHeight );
		}
		cImage.Sync();
		BeginDrag( &cData, cHotSpot, &cImage );
	}
	else
	{
		BeginDrag( &cData, cHotSpot, cSelRect );
	}
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void DirectoryView::HandleMessage( Message* pcMessage )
{
    switch( pcMessage->GetCode() )
	{
		case M_CLEAR:
		{
			Clear();
			break;
		}
		case M_ADD_ENTRY:
		{
//            std::string cName;
			const char* pzName;
			const struct stat* psStat;

			size_t nSize;


			int32 nCount = 0;
			pcMessage->GetNameInfo( "stat", NULL, &nCount );
			for ( int i = 0 ; i < nCount ; ++i )
			{
				if ( pcMessage->FindString( "name", &pzName, i ) != 0 )
					break;
	
				if ( pcMessage->FindData( "stat", T_ANY_TYPE, (const void**) &psStat, &nSize, i ) != 0 )
					break;










				try
				{
					FileRow* pcRow = new FileRow( m_pcIconBitmap, pzName, *psStat );

					const void* pIconData;
					if ( pcMessage->FindData( "small_icon", T_ANY_TYPE, &pIconData, &nSize, i ) == 0 )
						memcpy( pcRow->m_anIcon, pIconData, 16 * 16 * 4 );

					InsertRow( pcRow );
				}
				catch( std::exception& )
				{
				}
			}

			if( m_pcDirKeeper )
				m_pcDirKeeper->PostMessage( DirKeeper::M_ENTRIES_ADDED, m_pcDirKeeper, m_pcDirKeeper );
			break;
		}
		case M_UPDATE_ENTRY:
		{
			const char* pzName;
			const struct stat* psStat;
			size_t nSize;
			int32 nCount = 0;
			pcMessage->GetNameInfo( "stat", NULL, &nCount );

			for ( int i = 0 ; i < nCount ; ++i )
			{
				if ( pcMessage->FindString( "name", &pzName, i ) != 0 )
					break;

				if ( pcMessage->FindData( "stat", T_ANY_TYPE, (const void**) &psStat, &nSize, i ) != 0 )
					break;

				try
				{
					for ( ListView::const_iterator j = begin() ; j != end() ; ++j )
					{
						FileRow* pcRow = static_cast<FileRow*>(*j);
						if ( pcRow->m_sStat.st_ino == psStat->st_ino && pcRow->m_sStat.st_dev == psStat->st_dev )
						{
							const void* pIconData;
							pcRow->m_cName = pzName;
							pcRow->m_sStat = *psStat;
							if ( pcMessage->FindData( "small_icon", T_ANY_TYPE, &pIconData, &nSize, i ) == 0 )
							{
								memcpy( pcRow->m_anIcon, pIconData, 16 * 16 * 4 );
								InvalidateRow( j - begin(), ListView::INV_VISUAL );
							}
							break;
						}
//                        RemoveRow( j - begin() );
//                        InsertRow( pcRow );
					}
				}
				catch( std::exception& )
				{
				}
			}
			if( m_pcDirKeeper )
				m_pcDirKeeper->PostMessage( DirKeeper::M_ENTRIES_UPDATED, m_pcDirKeeper, m_pcDirKeeper );
			break;
		}
		case M_REMOVE_ENTRY:
		{
			dev_t nDevice;
			int64 nInode;

			int32 nCount = 0;
			pcMessage->GetNameInfo( "device", NULL, &nCount );

			for ( int i = 0 ; i < nCount ; ++i )
			{
				pcMessage->FindInt( "device", &nDevice, i );
				pcMessage->FindInt64( "node", &nInode, i );

				for ( ListView::const_iterator j = begin() ; j != end() ; ++j )
				{
					FileRow* pcRow = static_cast<FileRow*>(*j);
					if ( int64(pcRow->m_sStat.st_ino) == nInode &&
						pcRow->m_sStat.st_dev == nDevice )
					{
						delete RemoveRow( j - begin() );
						break;
					}
				}
			}
			if( m_pcDirKeeper )
				m_pcDirKeeper->PostMessage( DirKeeper::M_ENTRIES_REMOVED, m_pcDirKeeper, m_pcDirKeeper );
			break;
		}
		case M_LAYOUT:
		{
//			Layout();


			break;
		}
		case 125:
		{
			ReRead();
			break;
		}
		case M_CUT:
		case M_COPY:
		case M_PASTE:
		{
			Alert *aboutBox = new Alert("Manager v0.0.2", "Manager v0.0.2 :\n\nFile management utility for AtheOS\nWritten by Stephane Aubry (arbogad@free.fr)\n\nhttp://arbogad.free.fr/atheos/\n\nThis program is completely free (GPL or someting like that),\nshare it at will !", 0x00, "Close", NULL);
			aboutBox->Go(new Invoker());
			break;
		}
		default:
			ListView::HandleMessage( pcMessage );
	}
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void FileRow::Paint( const Rect& cFrame, View* pcView, uint nColumn,
                     bool bSelected, bool bHighlighted, bool bHasFocus )
{
	if ( nColumn == 0 )
	{
		pcView->Sync(); // Make sure the previous icon is rendered before we overwrite the bitmap
		memcpy( m_pcIconBitmap->LockRaster(), m_anIcon, 16*16*4 );
//		ConvertIcon( m_pcIconBitmap, m_anIcon, false );
		pcView->DrawBitmap( m_pcIconBitmap, Rect(0, 0, 15.0f, 15.0f ), cFrame );
		return;
	}
	font_height sHeight;
	pcView->GetFontHeight( &sHeight );

	if ( bHighlighted )
	{
		pcView->SetFgColor( 0, 50, 200 );
	}
//	else if ( bSelected )
//	{
//		pcView->SetFgColor( 0, 0, 0 );
//	}
	else
	{
		pcView->SetFgColor( 255, 255, 255 );
	}

	if ( nColumn != 6 )
		pcView->FillRect( cFrame );

	float vFontHeight = sHeight.ascender + sHeight.descender;
	float vBaseLine = cFrame.top + (cFrame.Height()+1.0f) / 2 - vFontHeight / 2 + sHeight.ascender;
	pcView->MovePenTo( cFrame.left + 3, vBaseLine );

	char  zBuffer[256];
	const char* pzString = zBuffer;

	switch( nColumn )
	{
		case 1: // name
			pzString = m_cName.c_str();
			break;
		case 2: // size
			if ( S_ISDIR( m_sStat.st_mode ) )
			{
				strcpy( zBuffer, "<DIR>" );
			}
			else
			{
				sprintf( zBuffer, "%ld", m_sStat.st_size );
			}
			break;
		case 3: // attributes
			for ( int i = 0 ; i < 10 ; ++i )
			{
				if ( m_sStat.st_mode & (1<<i) )
				{
					zBuffer[i] = "drwxrwxrwx"[9-i];
				}
				else
				{
					zBuffer[i] = '-';
				}
			}
			zBuffer[10] = '\0';
			break;
		case 4: // date
		{
			time_t nTime = m_sStat.st_ctime;
			strftime( zBuffer, 256, "%d/%b/%Y", localtime( &nTime ) );
			break;
		}
		case 5: // time
		{
			time_t nTime = m_sStat.st_ctime;
			strftime( zBuffer, 256, "%H:%M:%S", localtime( &nTime ) );
			break;
		}
		case 6: // name (for drag image)
			pzString = m_cName.c_str();
			break;
		default:
			printf( "Error: FileRow::Draw() - Invalid column %d\n", nColumn );
			return;
	}
/*
	if ( bHighlighted || (bSelected && bHasFocus) )
	{
		pcView->SetFgColor( 255, 255, 255 );
	}
	else
	{
		pcView->SetFgColor( 0, 0, 0 );
	}
*/
	if ( bHighlighted && nColumn == 1 )
	{
		pcView->SetFgColor( 255, 255, 255 );
		pcView->SetBgColor( 0, 50, 200 );
	}
	else if ( bSelected && nColumn == 1 )
	{
		pcView->SetFgColor( 255, 255, 255 );
		pcView->SetBgColor( 0, 0, 0 );
	}
	else
	{
		pcView->SetBgColor( 255, 255, 255 );
		pcView->SetFgColor( 0, 0, 0 );
	}

	if ( bSelected && nColumn == 1 )
	{
		Color32_s aColor = ( 0, 0, 0, 0 );
		Rect cRect = cFrame;
		cRect.right = cRect.left + pcView->GetStringWidth( pzString ) +          4;
		cRect.top = vBaseLine - sHeight.ascender - 1;
		cRect.bottom = vBaseLine + sHeight.descender + 1;
		pcView->FillRect( cRect, aColor );
	}
	pcView->DrawString( pzString );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void DirectoryView::MouseUp( const Point& cPosition, uint32 nButtons, Message* pcData )
{
	Highlight( 0, GetRowCount() - 1, false, false );
	
	// If this wasn't a drag, just leave now
	if ( pcData == NULL )
	{
		ListView::MouseUp( cPosition, nButtons, NULL );
		return;
	}
	
	StopScroll();

	const char* pzPath;
	if ( pcData->FindString( "file/path", &pzPath ) != 0 )
		return;

    if ( m_cPath == Path(pzPath).GetDir() )
		return;

	FileRow* pcRow;

	int nSel = HitTest( cPosition );
	if ( nSel != -1 )
	{
		pcRow = dynamic_cast<FileRow*>(GetRow(nSel));
	}
	else
	{
		pcRow = NULL;
	}

	Path cDstDir;

	if ( pcRow != NULL && S_ISDIR( pcRow->m_sStat.st_mode ) )
	{
		cDstDir = m_cPath;
		cDstDir.Append( pcRow->GetName().c_str() );
	}
	else
	{
		cDstDir = m_cPath;
	}

	std::vector<std::string> cSrcPaths;
	std::vector<std::string> cDstPaths;
	for ( int i = 0 ; pcData->FindString( "file/path", &pzPath, i ) == 0 ; ++i )
	{
		Path cSrcPath( pzPath );
		Path cDstPath = cDstDir;
		cDstPath.Append( cSrcPath.GetLeaf() );
		cSrcPaths.push_back( pzPath );
		cDstPaths.push_back( cDstPath.GetPath() );
	}

//	if( os::Application::GetInstance()->GetQualifiers() & os::QUAL_ALT )
//	{
//		/* Create links */
//		for( uint i = 0; i < cSrcPaths.size(); i++ )
//		{
//			symlink( cSrcPaths[i].c_str(), cDstPaths[i].c_str() );
//		}
//	} else
//		/* Start file copy */
		StartFileCopy( cDstPaths, cSrcPaths, Messenger( this ) );
//		StartFileCopy( cDstPaths, cSrcPaths, Messenger( this ), new os::Message( 125 ) );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void DirectoryView::MouseMove( const Point& cNewPos, int nCode, uint32 nButtons, Message* pcData )
{
	if ( pcData == NULL )
	{
		ListView::MouseMove( cNewPos, nCode, nButtons, NULL );
		return;
	}

	if ( nCode == MOUSE_OUTSIDE )
		return;

	if ( nCode == MOUSE_EXITED )
	{
		Highlight( 0, GetRowCount() - 1, false, false );
		StopScroll();
		return;
	}

	const char* pzPath;
	if ( pcData->FindString( "file/path", &pzPath ) != 0 )
		return;

	Rect cBounds = GetBounds();
	if ( cNewPos.y < cBounds.top + AUTOSCROLL_BORDER )
	{
		StartScroll( SCROLL_DOWN, false );
	}
	else if ( cNewPos.y > cBounds.bottom - AUTOSCROLL_BORDER )
	{
		StartScroll( SCROLL_UP, false );
	}
	else
	{
		StopScroll();
	}

	os::ListView::MouseMove( cNewPos, nCode, nButtons, pcData );

	int nSel = HitTest( cNewPos );
	if ( nSel != -1 )
	{
		FileRow* pcRow = dynamic_cast<FileRow*>(GetRow(nSel));
		if ( pcRow != NULL && S_ISDIR( pcRow->m_sStat.st_mode ) )
		{
			Path cRowPath = m_cPath;
			cRowPath.Append( pcRow->GetName().c_str() );

			if( !( cRowPath == Path( pzPath ) ) )
			{
				Highlight( nSel, true, true );
				return;
			}
		}
	}

	Highlight( 0, GetRowCount() - 1, false, false );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void DirectoryView::MouseDown( const Point& cPosition, uint32 nButtons )
{
	switch( nButtons )
	{
//		case 1:
//			break;
		case 2:
			// Open right-mouse-click PopupMenu
			PopupMenu( cPosition );
//			pcMenu->Open( ConvertToScreen( cPosition ) );
//			pcMenu->SetTargetForItems( this );
			break;
		default:
			View::MouseDown( cPosition, nButtons );
			break;
	}
	return;
}

void FileRow::AttachToView( View* pcView, int nColumn )
{
	if ( nColumn == 0 )
	{
		m_avWidths[0] = 16.0f;
		return;
	}

	char  zBuffer[256];
	const char* pzString = zBuffer;

	switch( nColumn )
	{
		case 1: // name
			pzString = m_cName.c_str();
			break;
		case 2: // size
			if ( S_ISDIR( m_sStat.st_mode ) )
			{
				strcpy( zBuffer, "<DIR>" );
			}
			else
			{
				sprintf( zBuffer, "%ld", m_sStat.st_size );
			}
			break;
		case 3: // attributes
			for ( int i = 0 ; i < 10 ; ++i )
			{
				if ( m_sStat.st_mode & (1 << i) )
				{
					zBuffer[i] = "drwxrwxrwx"[9-i];
				}
				else
				{
					zBuffer[i] = '-';
				}
			}
			zBuffer[10] = '\0';
			break;
		case 4: // date
		{
			time_t nTime = m_sStat.st_ctime;
			strftime( zBuffer, 256, "%d/%b/%Y", localtime( &nTime ) );
			break;
		}
		case 5: // time
		{
			time_t nTime = m_sStat.st_ctime;
			strftime( zBuffer, 256, "%H:%M:%S", localtime( &nTime ) );
			break;
		}
		case 6: // name (for drag image)
			pzString = m_cName.c_str();
			break;
		default:
			printf( "Error: FileRow::AttachToView() - Invalid column %d\n", nColumn );
			return;
	}
	m_avWidths[nColumn] = pcView->GetStringWidth( pzString ) + 5.0f;
}

void FileRow::SetRect( const Rect& cRect, int nColumn )
{
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

float FileRow::GetWidth( View* pcView, int nColumn )
{
	return( m_avWidths[nColumn] );























































}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

float FileRow::GetHeight( View* pcView )
{
	font_height sHeight;
	pcView->GetFontHeight( &sHeight );

	return( std::max( 16.0f - 3.0f, sHeight.ascender + sHeight.descender ) );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

bool FileRow::HitTest( View* pcView, const Rect& cFrame, int nColumn, Point cPos )
{
	if ( nColumn != 1 )
		return( true );

	font_height sHeight;
	pcView->GetFontHeight( &sHeight );
	float nFontHeight = sHeight.ascender + sHeight.descender;
	float nBaseLine = cFrame.top + (cFrame.Height()+1.0f) * 0.5f - nFontHeight / 2 + sHeight.ascender;
	Rect cRect( cFrame.left,
				 nBaseLine - sHeight.ascender - 1.0f,
				 cFrame.left + pcView->GetStringWidth( m_cName.c_str() ) + 4.0f,
				 nBaseLine + sHeight.descender + 1.0f );

	return( cRect.DoIntersect( cPos ) );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

bool FileRow::IsLessThan( const ListViewRow* pcOther, uint nColumn ) const
{
	const FileRow* pcRow = dynamic_cast<const FileRow*>(pcOther);
	if ( NULL == pcRow )
		return( false );

	if ( S_ISDIR( m_sStat.st_mode ) != S_ISDIR( pcRow->m_sStat.st_mode ) )
		return( S_ISDIR( m_sStat.st_mode ) );

	switch( nColumn )
	{
		case 0: // icon
		case 1: // name
			return( strcasecmp( m_cName.c_str(), pcRow->m_cName.c_str() ) < 0 );
		case 2: // size
			return( m_sStat.st_size < pcRow->m_sStat.st_size );
		case 3: // attributes
			return( strcasecmp( m_cName.c_str(), pcRow->m_cName.c_str() ) < 0 );
		case 4: // date
		case 5: // time
			return( m_sStat.st_mtime < pcRow->m_sStat.st_mtime );
		default:
//			printf( "Error: FileRow::IsLessThan() - Invalid column %d\n", nColumn );
			return( false );
	}
}






